@@ -51,6 +51,7 @@ struct _GSSDPResourceBrowserPrivate {
         GSSDPClient *client;
 
         char        *target;
+        GRegex      *target_regex;
 
         gushort      mx;
 
@@ -60,7 +61,7 @@ struct _GSSDPResourceBrowserPrivate {
 
         GHashTable  *resources;
                         
-        guint        timeout_id;
+        GSource     *timeout_src;
         guint        num_discovery;
 };
 
@@ -83,7 +84,7 @@ static guint signals[LAST_SIGNAL];
 typedef struct {
         GSSDPResourceBrowser *resource_browser;
         char                 *usn;
-        guint                 timeout_id;
+        GSource              *timeout_src;
 } Resource;
 
 /* Function prototypes */
@@ -220,6 +221,8 @@ gssdp_resource_browser_dispose (GObject *object)
         }
 
         clear_cache (resource_browser);
+
+        G_OBJECT_CLASS (gssdp_resource_browser_parent_class)->dispose (object);
 }
 
 static void
@@ -229,9 +232,14 @@ gssdp_resource_browser_finalize (GObject *object)
 
         resource_browser = GSSDP_RESOURCE_BROWSER (object);
 
+        if (resource_browser->priv->target_regex)
+                g_regex_unref (resource_browser->priv->target_regex);
+
         g_free (resource_browser->priv->target);
 
         g_hash_table_destroy (resource_browser->priv->resources);
+
+        G_OBJECT_CLASS (gssdp_resource_browser_parent_class)->finalize (object);
 }
 
 static void
@@ -430,6 +438,11 @@ void
 gssdp_resource_browser_set_target (GSSDPResourceBrowser *resource_browser,
                                    const char           *target)
 {
+        char *pattern;
+        char *version;
+        char *version_pattern;
+        GError *error;
+
         g_return_if_fail (GSSDP_IS_RESOURCE_BROWSER (resource_browser));
         g_return_if_fail (target != NULL);
         g_return_if_fail (!resource_browser->priv->active);
@@ -437,6 +450,34 @@ gssdp_resource_browser_set_target (GSSDPResourceBrowser *resource_browser,
         g_free (resource_browser->priv->target);
         resource_browser->priv->target = g_strdup (target);
 
+        if (resource_browser->priv->target_regex)
+                g_regex_unref (resource_browser->priv->target_regex);
+
+        version_pattern = "[0-9]+";
+        /* Make sure we have enough room for version pattern */
+        pattern = g_strndup (target,
+                             strlen (target) + strlen (version_pattern));
+
+        version = g_strrstr (pattern, ":") + 1;
+        if (version != NULL &&
+            g_regex_match_simple (version_pattern, version, 0, 0)) {
+                strcpy (version, version_pattern);
+        }
+
+        error = NULL;
+        resource_browser->priv->target_regex = g_regex_new (pattern,
+                                                            0,
+                                                            0,
+                                                            &error);
+        if (error) {
+                g_warning ("Error compiling regular expression '%s': %s",
+                           pattern,
+                           error->message);
+
+                g_error_free (error);
+        }
+
+        g_free (pattern);
         g_object_notify (G_OBJECT (resource_browser), "target");
 }
 
@@ -540,16 +581,23 @@ static gboolean
 resource_expire (gpointer user_data)
 {
         Resource *resource;
+        char *usn;
 
         resource = user_data;
-        
+
+        /* Steal the USN pointer from the resource as we need it for the signal
+         * emission.
+         */
+        usn = resource->usn;
+        resource->usn = NULL;
+
+        g_hash_table_remove (resource->resource_browser->priv->resources, usn);
+
         g_signal_emit (resource->resource_browser,
                        signals[RESOURCE_UNAVAILABLE],
                        0,
-                       resource->usn);
-
-        g_hash_table_remove (resource->resource_browser->priv->resources,
-                             resource->usn);
+                       usn);
+        g_free (usn);
 
         return FALSE;
 }
@@ -564,8 +612,9 @@ resource_available (GSSDPResourceBrowser *resource_browser,
         gboolean was_cached;
         guint timeout;
         GList *locations;
+        GMainContext *context;
 
-        usn = soup_message_headers_get (headers, "USN");
+        usn = soup_message_headers_get_one (headers, "USN");
         if (!usn)
                 return; /* No USN specified */
 
@@ -573,7 +622,7 @@ resource_available (GSSDPResourceBrowser *resource_browser,
         resource = g_hash_table_lookup (resource_browser->priv->resources, usn);
         if (resource) {
                 /* Remove old timeout */
-                g_source_remove (resource->timeout_id);
+                g_source_destroy (resource->timeout_src);
 
                 was_cached = TRUE;
         } else {
@@ -591,7 +640,7 @@ resource_available (GSSDPResourceBrowser *resource_browser,
         }
 
         /* Calculate new timeout */
-        header = soup_message_headers_get (headers, "Cache-Control");
+        header = soup_message_headers_get_one (headers, "Cache-Control");
         if (header) {
                 GSList *list;
                 int res;
@@ -622,7 +671,7 @@ resource_available (GSSDPResourceBrowser *resource_browser,
         } else {
                 const char *expires;
 
-                expires = soup_message_headers_get (headers, "Expires");
+                expires = soup_message_headers_get_one (headers, "Expires");
                 if (expires) {
                         SoupDate *soup_exp_time;
                         time_t exp_time, cur_time;
@@ -653,9 +702,16 @@ resource_available (GSSDPResourceBrowser *resource_browser,
                 }
         }
 
-        resource->timeout_id = g_timeout_add_seconds (timeout,
-                                                      resource_expire,
-                                                      resource);
+        resource->timeout_src = g_timeout_source_new_seconds (timeout);
+	g_source_set_callback (resource->timeout_src,
+                               resource_expire,
+			       resource, NULL);
+
+        context = gssdp_client_get_main_context
+                (resource_browser->priv->client);
+	g_source_attach (resource->timeout_src, context);
+
+        g_source_unref (resource->timeout_src);
 
         /* Only continue with signal emission if this resource was not
          * cached already */
@@ -665,11 +721,11 @@ resource_available (GSSDPResourceBrowser *resource_browser,
         /* Build list of locations */
         locations = NULL;
 
-        header = soup_message_headers_get (headers, "Location");
+        header = soup_message_headers_get_one (headers, "Location");
         if (header)
                 locations = g_list_append (locations, g_strdup (header));
 
-        header = soup_message_headers_get (headers, "AL");
+        header = soup_message_headers_get_one (headers, "AL");
         if (header) {
                 /* Parse AL header. The format is:
                  * <uri1><uri2>... */
@@ -714,7 +770,7 @@ resource_unavailable (GSSDPResourceBrowser *resource_browser,
 {
         const char *usn;
 
-        usn = soup_message_headers_get (headers, "USN");
+        usn = soup_message_headers_get_one (headers, "USN");
         if (!usn)
                 return; /* No USN specified */
 
@@ -722,12 +778,24 @@ resource_unavailable (GSSDPResourceBrowser *resource_browser,
         if (!g_hash_table_lookup (resource_browser->priv->resources, usn))
                 return;
 
+        g_hash_table_remove (resource_browser->priv->resources, usn);
+
         g_signal_emit (resource_browser,
                        signals[RESOURCE_UNAVAILABLE],
                        0,
                        usn);
+}
 
-        g_hash_table_remove (resource_browser->priv->resources, usn);
+static gboolean
+check_target_compat (GSSDPResourceBrowser *resource_browser,
+                     const char           *st)
+{
+        return strcmp (resource_browser->priv->target,
+                       GSSDP_ALL_RESOURCES) == 0 ||
+               g_regex_match (resource_browser->priv->target_regex,
+                              st,
+                              0,
+                              NULL);
 }
 
 static void
@@ -736,12 +804,11 @@ received_discovery_response (GSSDPResourceBrowser *resource_browser,
 {
         const char *st;
 
-        st = soup_message_headers_get (headers, "ST");
+        st = soup_message_headers_get_one (headers, "ST");
         if (!st)
                 return; /* No target specified */
 
-        if (strcmp (resource_browser->priv->target, GSSDP_ALL_RESOURCES) != 0 &&
-            strcmp (resource_browser->priv->target, st) != 0)
+        if (!check_target_compat (resource_browser, st))
                 return; /* Target doesn't match */
 
         resource_available (resource_browser, headers);
@@ -753,15 +820,14 @@ received_announcement (GSSDPResourceBrowser *resource_browser,
 {
         const char *header;
 
-        header = soup_message_headers_get (headers, "NT");
+        header = soup_message_headers_get_one (headers, "NT");
         if (!header)
                 return; /* No target specified */
 
-        if (strcmp (resource_browser->priv->target, GSSDP_ALL_RESOURCES) != 0 &&
-            strcmp (resource_browser->priv->target, header) != 0)
+        if (!check_target_compat (resource_browser, header))
                 return; /* Target doesn't match */
 
-        header = soup_message_headers_get (headers, "NTS");
+        header = soup_message_headers_get_one (headers, "NTS");
         if (!header)
                 return; /* No announcement type specified */
 
@@ -818,7 +884,7 @@ resource_free (gpointer data)
 
         g_free (resource->usn);
 
-        g_source_remove (resource->timeout_id);
+        g_source_destroy (resource->timeout_src);
 
         g_slice_free (Resource, resource);
 }
@@ -859,7 +925,7 @@ send_discovery_request (GSSDPResourceBrowser *resource_browser)
         message = g_strdup_printf (SSDP_DISCOVERY_REQUEST,
                                    resource_browser->priv->target,
                                    resource_browser->priv->mx,
-                                   g_get_application_name ());
+                                   g_get_application_name () ?: "");
 
         _gssdp_client_send_message (resource_browser->priv->client,
                                     NULL,
@@ -881,7 +947,7 @@ discovery_timeout (gpointer data)
         resource_browser->priv->num_discovery += 1;
 
         if (resource_browser->priv->num_discovery >= MAX_DISCOVERY_MESSAGES) {
-                resource_browser->priv->timeout_id = 0;
+		resource_browser->priv->timeout_src = NULL;
                 resource_browser->priv->num_discovery = 0;
 
                 return FALSE;
@@ -893,24 +959,33 @@ discovery_timeout (gpointer data)
 static void
 start_discovery (GSSDPResourceBrowser *resource_browser)
 {
+        GMainContext *context;
+
         /* Send one now */
         send_discovery_request (resource_browser);
 
         /* And schedule the rest for later */
         resource_browser->priv->num_discovery = 1;
-        resource_browser->priv->timeout_id =
-                                g_timeout_add (DISCOVERY_FREQUENCY,
-                                               discovery_timeout,
-                                               resource_browser);
+        resource_browser->priv->timeout_src =
+                g_timeout_source_new (DISCOVERY_FREQUENCY);
+	g_source_set_callback (resource_browser->priv->timeout_src,
+			       discovery_timeout,
+			       resource_browser, NULL);
+
+        context = gssdp_client_get_main_context
+                (resource_browser->priv->client);
+	g_source_attach (resource_browser->priv->timeout_src, context);
+
+        g_source_unref (resource_browser->priv->timeout_src);
 }
 
 /* Stops the sending of discovery messages */
 static void
 stop_discovery (GSSDPResourceBrowser *resource_browser)
 {
-        if (resource_browser->priv->timeout_id) {
-                g_source_remove (resource_browser->priv->timeout_id);
-                resource_browser->priv->timeout_id = 0;
-                resource_browser->priv->num_discovery = 0;
-        }
+	if (resource_browser->priv->timeout_src) {
+		g_source_destroy (resource_browser->priv->timeout_src);
+		resource_browser->priv->timeout_src = NULL;
+		resource_browser->priv->num_discovery = 0;
+	}
 }
